#########################################################

# Description: ASM implementation (only the shape model, no image information)
# The basic procedures are:
# (1) align all models together, get the mean shape
# (2) do PCA on the covariance matrix to get "largest eigenvectors" and eigenvalues
# (3) Given a new shape, refine it by projecting to the PCA space, which will remove small variances

# Usage: check out Test() method (may need to change some global variables)

# Author: Shaoting Zhang, shaoting@cs.rutgers.edu

#########################################################

import os
import re
import copy
import numpy
import Utility

### some global variables, such as the location of files

# dataRoot: the name of a folder which may contain subfolders and Off files
global dataRoot

# Need connectivity of any reference mesh (all meshes share the same connectiviy)
global referenceMesh

# directory to save the result, debug purpose
global resultDir

# alignMode = 1 means similarity transform; 
# alignMode = 2 means rigid transformation;
# other values mean that no alignment is applied
global alignMode

# point set of mean shape
global meanShape 

# "largest" eigenvectors, i.e. modes
global P

# square root of largest eigenvalues, i.e. standard deviation
global std

# 2D or 3D
global dimension

#########################################################################
### Provide interface to set data value
def SetDataRoot(rootTmp):
    global dataRoot
    dataRoot = rootTmp
    
def SetReferenceMesh(meshTmp):
    global referenceMesh
    referenceMesh = meshTmp
    
def SetResultDir(dirTmp):
    global resultDir
    resultDir = dirTmp
    
def SetAlignMode(modeTmp):
    global alignMode
    alignMode = modeTmp
    
def SetDimension(dimensionTmp):
    global dimension
    dimension = dimensionTmp
    
#########################################################################
### interface for segmentation tools
def PreProcess(dataRootTmp=r'.\data', 
referenceMeshTmp=r'.\data\set1\1.off', 
alignModeTmp=1, # Similarity transformation
dimensionTmp=3, # 3D case
maskTmp='.off'):
    global meanShape
    global P
    global std
    global dataRoot
    global referenceMesh
    global resultDir
    global alignMode
    global dimension
    
    ### for 3D case
    SetResultDir(r'.\asm_mode')
    SetDataRoot(dataRootTmp)
    SetReferenceMesh(referenceMeshTmp)
    SetAlignMode(alignModeTmp)
    SetDimension(dimensionTmp)
    meshes = LoadAllMeshes(maskTmp)
    alignedMeshes = AlignTrainingSet(meshes, meshes[0])
    GetMeanShape(alignedMeshes)
    alignedMeshes = AlignTrainingSet(meshes, meanShape)
    GetMeanShape(alignedMeshes)
    CaptureStatistics(alignedMeshes)
	
    # save major shape variance as VTK files. Can use Animation_TK.py to visualize them in sequence
    SaveVarianceAsVTK()
    
    # This is for 2D data. Save everything to txt files
    #Save2DPointSet2TXT(meanShape, r'C:\test\mean.txt')

# inputs are names of two iso files. 
# Since this module relies on off dataset, we need to convert data first.
def ASMRefinement(resultSurface='noise_shape.off',refinedSurface='refined_shape.off'):
    currentShape = LoadOneMesh(resultSurface)
    refinedShape = DeformationUsingShapes(currentShape)
    SavePointSet2Off(refinedShape, refinedSurface)

# recursively load all off files in "dataRoot"
# return value is a list. Each element is a point set (a list containing all vertices) looks like:
#    v1      v2       v3  ....   v4
# [[0,0,0],[1,1,1],[2,2,2],...,[n,n,n]]
def LoadAllMeshes(mask):
    global dataRoot
    global dimension
    
    amount = 0
    meshes = []
    for root, dirs, files in os.walk(dataRoot): 
        if [] != dirs: 
            continue
        for fileid in range(len(files)):
            if re.search(mask, files[fileid]) != None:
                filename = root + '\\' + files[fileid]
                curMesh = LoadOneMesh(filename)
                meshes.append(curMesh)
                amount = amount + 1
    print 'Loaded all meshes (.off files) \n'
    return meshes
    
def LoadOneMesh(filename):
    global dimension
    
    file = open(filename, 'r')
    if dimension == 3: # in 3D case, the first line will be "OFF"
        file.readline()
    size = file.readline()
    size = size.split()
    mesh = [[0]*dimension] * int(size[0])
    for vertexid in range(int(size[0])):
        curLine = file.readline()
        curLine = curLine.split()
        curLine = map(float, curLine) # cast string to int
        mesh[vertexid] = curLine
    file.close()
    return mesh
    
def GetMeanShape(meshes):
    global meanShape
    global dimension
    
    amount = len(meshes)
    meshSize = len(meshes[0])
    meanShape = [[0]*dimension] * meshSize
    for meshid in range(amount):
        curMesh = meshes[meshid]
        meanShape = AddMesh(meanShape, curMesh)
            
    for vertexid in range(meshSize):
        meanShape[vertexid] = Utility.scale(1.0/float(amount), meanShape[vertexid])
    print 'Computed the mean shape \n'
    return meanShape
    
# Rotate, scale and translate (similarity transform) each shape to align with meshSrc
# meshSrc can be the first mesh in the dataset, or the mean shape
# Return value is all meshes after transformation
def AlignTrainingSet(meshes, meshDes):
    amount = len(meshes)
    alignedMeshes = copy.copy(meshes)
    for meshid in range(amount):
        meshSrc = meshes[meshid]
        if alignMode == 1:
            alignedMeshes[meshid] = Utility.SimilarityTransformationGeneral(meshSrc, meshDes, dimension)
        elif alignMode == 2:
            alignedMeshes[meshid] = Utility.RigidTransformationGeneral(meshSrc, meshDes,dimension)
        else:
            alignedMeshes[meshid] = meshSrc
    print 'Meshes were aligned \n'
    return alignedMeshes

# Compute variance of main modes
# The notation is similar to "Active Shape Models - Their Training and Application"
# Appendix B, faster way to compute PCA when sample size is smaller than the dimension
# The return values are the "largest" eigenvectors (P) and eigenvalues (std)
def CaptureStatistics(meshes, varianceThreshold = .9):
    global meanShape
    global P
    global std
    global dimension
    
    N = len(meshes)
    meshSize = len(meshes[0])
    Dt = numpy.zeros([N, meshSize*dimension])
    xbar = Mesh2Vector(meanShape)
    for meshid in range(N):
        x = Mesh2Vector(meshes[meshid])
        dx = numpy.subtract(x, xbar)
        Dt[meshid] = numpy.transpose(dx)
    D = numpy.transpose(Dt);
    T = numpy.dot(Dt, D)
    T = numpy.divide(T, N)
    u, lam, uh = numpy.linalg.svd(T)
    lamSum = numpy.sum(lam)
    variance = 0
    index = 0
    P = []
    std = []
    u = numpy.transpose(u)
    while variance < varianceThreshold:
        variance = variance + lam[index] / lamSum
        pi = numpy.dot(D, u[index])
        pi = numpy.divide(pi, numpy.sqrt(lam[index]*N))
        P.append(pi)
        std.append(numpy.sqrt(lam[index]))
        index = index + 1
        print variance
    
# ASM paper, Sec4.1, calculating a suggested movement for each model point
# Given a current shape and the mean shape, we first align (rigid, similarity, etc)
# the current shape to the mean shape; then compute the difference (dx) between two shapes;
# refine dx according to statistical information (P, std) and compute db;
# finally update the shape and transform it back using the opposite transformation
def DeformationUsingShapes(currentShape):
    global meanShape
    global P
    global std
    
    # 1. align the current shape to the mean shape
    if alignMode == 1:
        [rotation,translation,scaling] = Utility.EstimateSimilarityTransformationGeneral(currentShape, meanShape,dimension)
        alignedMesh = Utility.ApplySimilarityTransformation(currentShape,rotation,translation,scaling)
    elif alignMode == 2:
        [rotation,translation] = Utility.EstimateRigidTransformationGeneral(currentShape,meanShape,dimension)
        alignedMesh = Utility.ApplyRigidTransformation(currentShape,rotation,translation)
    else:
        alignedMesh = currentShape
    # 2. project to PCA space, and map it back, to remove unlikely deformation
    xbar = Mesh2Vector(meanShape)
    x = Mesh2Vector(alignedMesh)
    dx = numpy.subtract(x, xbar)
    db = numpy.dot(P, dx)
    # truncate db if it's larger than 2.5 * standard deviation (variance is too large)
    for db_id in range(len(db)):
        if abs(db[db_id])>2.5*abs(std[db_id]):
            db[db_id] = numpy.sign(db[db_id]) * 2.5 * abs(std[db_id])
    Pdb = numpy.dot(numpy.transpose(P), db)
    x_approximate = numpy.add(xbar, Pdb)
    PCAMesh = Vector2Mesh(x_approximate)
    # 3. transform it back, negative rotation, negative translation, reciprocal of scaling
    if alignMode == 1:
        resultMesh = Utility.ApplyReverseSimilarityTransformation(PCAMesh,rotation,translation,scaling)
    elif alignMode == 2:
        resultMesh = Utility.ApplyReverseRigidTransformation(PCAMesh,rotation,translation)
    else:
        resultMesh = PCAMesh
    return resultMesh
    
# save significant modes with variance changing from -2*std to 2*std
# Generate one subfolder for each mode
def SaveVarianceAsVTK(numberOfMeshes = 9):
    global meanShape
    global P
    global std
    
    modeSize = len(std)
    xbar = Mesh2Vector(meanShape)
    for modeid in range(modeSize):
        curP = P[modeid]
        curStd = std[modeid]
        stepSize = 6.0 * curStd / float(numberOfMeshes-1)
        folderName = resultDir + '\\mode' + str(modeid+1)
        try:
            os.mkdir(folderName)
        except os.error: # in case the directory is already there
            pass
        for i in range(numberOfMeshes):
            bk = -3.0*curStd + stepSize * i
            Pb = numpy.dot(P[modeid], bk)
            Pb = numpy.reshape(Pb, numpy.shape(xbar))
            x = numpy.add(xbar, Pb)
            mesh = Vector2Mesh(x)
            savedNameOff = resultDir + '\\mode' + str(modeid+1) + '\\' + str(i+1) + '.off'
            savedNameVTK = resultDir + '\\mode' + str(modeid+1) + '\\' + str(i+1) + '.vtk'
            SavePointSet2Off(mesh, savedNameOff)
            Utility.off2vtk(savedNameOff, savedNameVTK)
            
# save all meshes into VTK files
def SaveMeshesAsVTK(meshes):
    for meshid in range(len(meshes)):
        savedNameOff = resultDir + '\\' + str(meshid+1) + '.off'
        savedNameVTK = resultDir + '\\' + str(meshid+1) + '.vtk'
        SavePointSet2Off(meshes[meshid], savedNameOff)
        Utility.off2vtk(savedNameOff, savedNameVTK)
    
#########################################################################
### helper methods, private...

def Mesh2Vector(mesh):
    global dimension
    
    meshSize = len(mesh)
    result = numpy.zeros((meshSize*dimension, 1))
    for i in range(meshSize):
        curPt = mesh[i]
        for j in range(dimension):
            result[i*dimension + j, 0] = curPt[j]
    return result
    
def Vector2Mesh(vector):
    global dimension
    
    vecSize = numpy.size(vector)
    meshSize = vecSize / dimension
    result = [[0]*dimension] * meshSize
    for vertexid in range(meshSize):
        if dimension == 3:
            result[vertexid] = [vector[vertexid*3,0], vector[vertexid*3+1,0], vector[vertexid*3+2,0]]
        else:
            result[vertexid] = [vector[vertexid*2,0], vector[vertexid*2+1,0]]
    return result
    
def AddMesh(mesh1, mesh2):
    meshSize = len(mesh1)
    result = copy.copy(mesh1)
    for vertexid in range(meshSize):
        result[vertexid] = Utility.addvec(mesh1[vertexid], mesh2[vertexid])
    return result

def SubMesh(mesh1, mesh2):
    meshSize = len(mesh1)
    result = copy.copy(mesh1)
    for vertexid in range(meshSize):
        result[vertexid] = Utility.subvec(mesh1[vertexid], mesh2[vertexid])
    return result
    
# save 3D point set (a list in python, each element is a list with 3 coordinates) into Off file
# Since there is no connectivity information in the point set, a reference file is necessary
# fileDes: the name of the saved file
def SavePointSet2Off(pointSet, fileDes):
    src = open(referenceMesh, 'r')
    des = open(fileDes, 'w')
    src.readline()
    size = src.readline()
    des.write('OFF\n')
    des.write(size)
    size = size.split()
    for i in range(int(size[0])):
        src.readline()
        curPt = pointSet[i]
        des.write(str(curPt[0]) + ' ' + str(curPt[1]) + ' ' + str(curPt[2]) + '\n')
        
    for i in range(int(size[1])):
        des.write(src.readline())
    src.close()
    des.close()
    print 'Saved point set to Off file: ' + fileDes + '\n'
    return
    
# save 2D point set into txt file
def Save2DPointSet2TXT(pointSet, fileDes):
    size = numpy.shape(pointSet)
    des = open(fileDes, 'w')
    #des.write(str(size[0]))
    for i in range(int(size[0])):
        curPt = pointSet[i]
        des.write(str(curPt[0]) + ' ' + str(curPt[1]) + ' ' + '\n')
        
    des.close()
    print 'Saved point set to Off file: ' + fileDes + '\n'
    return