#########################################################

# Description: anomaly detection using sparse coding
# The basic assumption is that each input shape(or landmarks) can be represented a sparse linear
# combination of training data. Futuremore, there might be errors in the input. Some of the detected
# points can be wrong, but the error points should also be sparse.

# The basic procedures are 
# (1) align data (Currently I did similarity transformation to get the mean shape,
# then use rigid transformation to align everything to this mean shape)
# (2) Prepare data matrix (each column is one mesh (or landmarks from one mesh) from the training data)
# (3) Given detected landmarks, generate a vector for CVXMOD, and then solve the system

# Note: need to install cvxmod (http://cvxmod.net/)

# Author: Shaoting Zhang, shaoting@cs.rutgers.edu

#########################################################

import os
import re
import copy
import numpy
import Utility

from cvxmod import *
from cvxmod.atoms import *
from cvxmod.sets import *

import ASMtools # need to save/load meshes, align shapes, etc. re-use code.

global landmarkList         # a list of landmark IDs, so we only use a subset of the mesh
global landmarkMatrix       # landmark matrix, each column represents one instance (landmark coordinates from one training mesh)
global meshMatrix           # mesh matrix, each column represents one mesh
global partitionMatrixList  # list of matrices, each matrix represents one partition's information (all meshes)
global meanShapeVectorList  # list of vectors, each vector denotes one partition of the mean mesh
global meanShape            # mesh of the mean shape
global alignedMeshes        # meshes after alignment
global meshNum              # number of instances (#meshes)
global vertexNum            # number of vertices of one mesh
global dimension            # 2D or 3D
global landmarkNum          # number of landmarks
global rotation             # when applying transformation matrix to the input mesh/landmarks, we need to keep a record
global translation
global scaling
global alignMode            # 1 means similarity transformation, 2 means rigid transformation

def SetLandmarkList(listTmp):
    global landmarkList
    landmarkList = listTmp
    
def SetAlignMode(modeTmp):
    global alignMode
    alignMode = modeTmp
    
def SetDimension(dimensionTmp):
    global dimension
    dimension = dimensionTmp

def AlignMeshes(dataRoot, referenceMesh, fileNameMask, dimensionTmp):
    global meanShape
    global alignedMeshes
    global meshNum
    global vertexNum
    
    SetDimension(dimensionTmp)
    ASMtools.SetDimension(dimensionTmp)
    ASMtools.SetDataRoot(dataRoot)
    ASMtools.SetReferenceMesh(referenceMesh)
    meshes = ASMtools.LoadAllMeshes(fileNameMask)
    
    ASMtools.SetAlignMode(1) # similarity transformation to find mean shape
    alignedMeshes = ASMtools.AlignTrainingSet(meshes, meshes[0])
    meanShape = ASMtools.GetMeanShape(alignedMeshes)
    alignedMeshes = ASMtools.AlignTrainingSet(meshes, meanShape)
    
    size = numpy.shape(alignedMeshes) # [#meshes, #vertices of one mesh, dimension (3)]
    meshNum = size[0]
    vertexNum = size[1]
    dimension = size[2]

# align meshes, and then generate data matrix (A)
def PreProcess(landmarkListTmp=range(30), dataRoot=r'.\data', 
referenceMesh=r'.\data\set1\1.off',
fileNameMask='.off', dimensionTmp=3):

    global landmarkList
    global landmarkMatrix
    global meshMatrix
    
    global meanShape
    global alignedMeshes
    global meshNum
    global landmarkNum
    global dimension
    global vertexNum
    
    AlignMeshes(dataRoot, referenceMesh, fileNameMask, dimensionTmp)
    
    SetLandmarkList(landmarkListTmp)
    landmarkNum = numpy.size(landmarkList)
    n_row = landmarkNum
    n_col = meshNum
    
    landmarkMatrix = rand(n_row*dimension, n_col)
    
    meshMatrix = rand(vertexNum*dimension, n_col)
    
    for i in range(meshNum):
        curMesh = numpy.transpose(alignedMeshes[i])
        for j in range(dimension):
            landmarkMatrix[n_row*j:n_row*j+n_row,i] = list(numpy.take(curMesh[j], landmarkList))
            meshMatrix[vertexNum*j:vertexNum*j+vertexNum,i] = list(curMesh[j])

# It is for the initialization (landmark) stage
# srcLMSet and destLMSet is a data structure in segmentation.py (LiverSegmentation)
# We need to convert it to CVXMOD matrix first
def DetectedLM2InputVector(srcLMSet, destLMSet):
    global dimension
    global landmarkNum
    
    global rotation
    global translation
    global scaling
    
    # transform the input landmarks to the mean shape
    if alignMode == 1:
        [rotation,translation,scaling] = Utility.EstimateSimilarityTransformationGeneral(destLMSet, srcLMSet, dimension)
        destLMSetTmp = Utility.ApplySimilarityTransformation(destLMSet,rotation,translation,scaling)
    else:
        [rotation,translation] = Utility.EstimateRigidTransformationGeneral(destLMSet, srcLMSet, dimension)
        destLMSetTmp = Utility.ApplyRigidTransformation(destLMSet,rotation,translation)
        
    inputVector = rand(landmarkNum*dimension, 1)
    
    # convert this point set into a vector, which can be used by CVXMOD
    for i in range(len(destLMSet)):
        for j in range(dimension):
            inputVector[landmarkNum*j + i] = destLMSetTmp[i][j]
    return inputVector

# It is for the deformation stage
# need to transform the whole shape (or part of it) to the mean shape,
# and store it as a CVXMOD matrix/vector
def DetectedShape2InputVector(curMesh):
    global dimension
    global landmarkList
    global landmarkNum
    global meanShape
    
    global rotation
    global translation
    global scaling
    
    # transform the input landmarks to the mean shape)
    if alignMode == 1:
        [rotation,translation,scaling] = Utility.EstimateSimilarityTransformationGeneral(curMesh, meanShape,dimension)
        destLMSetTmp = Utility.ApplySimilarityTransformation(curMesh,rotation,translation,scaling)
    else:
        [rotation,translation,scaling] = Utility.EstimateRigidTransformationGeneral(curMesh, meanShape,dimension)
        destLMSetTmp = Utility.ApplyRigidTransformation(curMesh,rotation,translation,scaling)
    
    inputVector = rand(landmarkNum*dimension, 1)
    
    destLMSetTmp = numpy.transpose(destLMSetTmp)
    for j in range(dimension):
        inputVector[landmarkNum*j:landmarkNum*j+landmarkNum] = list(numpy.take(destLMSetTmp[j], landmarkList))
    
    return inputVector

# This method can be used in both initialization and deformation steps
# landmarkMatrix - A, inputVec - L
# min_{T,f,x}[|Ax - TL - f|_2 + alpha |x|_1 + beta |f|_1]
# A is a n by m matrix, where n = landmarkNum x dimension, m = meshNum;
# x is m by 1 sparse vector, L is n by 1 (detected landmarks), f is n by 1 sparse vector (error detection)
def SparseRepresentation(srcLMSet, destLMSet, needShapeResult, alpha, beta):
    global landmarkMatrix
    global meshMatrix
    global meshNum
    global dimension
    global landmarkNum
    
    global rotation
    global translation
    global scaling
    
    # step 1, fix x and f, optimize T
    if srcLMSet == []: # for shape
        transformedDetectedLandmark = DetectedShape2InputVector(destLMSet) # now destLMSet = current shape (or part of the shape)
    else: # for landmark
        transformedDetectedLandmark = DetectedLM2InputVector(srcLMSet, destLMSet)

    # step 2, fix T, optimize x and f
    x = optvar('x', meshNum)
    error = optvar('f', landmarkNum*dimension)
    p = problem(minimize(norm2(landmarkMatrix*x-transformedDetectedLandmark - error) + alpha*norm1(x) + beta*norm1(error)))
    #p = problem(minimize(norm2(landmarkMatrix*x-transformedDetectedLandmark) + alpha*norm1(x)))
    p.solve()
    
    print error.value
    #print x.value
    
    refinedLM = []
    
    error_value = error.value
    error_array = map(float,(numpy.abs(error_value)))
    #error_array = 0
    
    LMVec = numpy.array((landmarkMatrix*x).value)
    refinedLM = Vector2Mesh(LMVec)
    if alignMode == 1:
        refinedLM = Utility.ApplyReverseSimilarityTransformation(refinedLM,rotation,translation,scaling)
    else:
        refinedLM = Utility.ApplyReverseRigidTransformation(refinedLM,rotation,translation)
    
    
    if needShapeResult == True: # compute a sparse linear combination of meshes
        meshFromSparseCoding = numpy.array(meshMatrix*x.value)
        
        meshFromSparseCoding = numpy.array(meshFromSparseCoding)
        meshFromSparseCoding = Vector2Mesh(meshFromSparseCoding) 
        
        if alignMode == 1:
            meshFromSparseCoding = Utility.ApplyReverseSimilarityTransformation(meshFromSparseCoding,rotation,translation,scaling)
        else:
            meshFromSparseCoding = Utility.ApplyReverseRigidTransformation(meshFromSparseCoding,rotation,translation)
    else:
        meshFromSparseCoding = []
    
    return [error_array, refinedLM, meshFromSparseCoding]
    
# Initialization for 2D application (e.g., lung localization task in our CVPR paper). 
# Remove one outlier each time if the error is large (>100)
def SparseRepresentationBasedInitialization_2D(srcLMSet, destLMSet, 
needShapeResult = True, 
alpha = 50, beta = 0.17, iterationNum = 5, threshold = 100,
landmarkListTmp=[], 
dataRoot=r'placeholder',
referenceMesh=r'placeholder',
fileNameMask='placeholder',
dimensionTmp = 2):

    global landmarkList
    global landmarkNum

    SetAlignMode(1) 
    
    for i in range(iterationNum):
        [error_array, refinedLM, meshFromSparseCoding] = SparseRepresentation(srcLMSet, destLMSet, needShapeResult, alpha, beta)

        error_max = numpy.max(error_array)
        prunedID = numpy.where(error_array == error_max)
        if error_array[prunedID[0]]>threshold and landmarkNum>3:
            prunedID = prunedID[0]%landmarkNum
            landmarkList.pop(prunedID)
            srcLMSet.pop(prunedID)
            destLMSet.pop(prunedID)
        else:
            break
            
        landmarkNum = numpy.size(landmarkList)
        PreProcess(landmarkList, dataRoot, referenceMesh, fileNameMask, dimensionTmp)
    
    SetAlignMode(1) # similarity transformation
    [error_array, refinedLM, meshFromSparseCoding] = SparseRepresentation(srcLMSet, destLMSet, needShapeResult, alpha, beta)

    return [refinedLM, meshFromSparseCoding]

# Initialization for 3D application (e.g., initialization step in the liver segmentation task in our CVPR paper)
def SparseRepresentationBasedInitialization_3D(srcLMSet, destLMSet, initSurface, needShapeResult = True, alpha = 50, beta = 0.17):
    SetAlignMode(1)
    [error_array, refinedLM, meshFromSparseCoding] = SparseRepresentation(srcLMSet, destLMSet, needShapeResult, alpha, beta)
    initSurface_off = initSurface + '.off'
    ASMtools.SavePointSet2Off(meshFromSparseCoding, initSurface_off)
    Utility.off2iso(initSurface_off, initSurface)

# A general interface which can be used during deformation (adaptively)
# Give a list of IDs, the current shape is refined by applying sparse analysis on points of IDs
# inputs are names of two iso files (current surface name and refined surface name)
# Since this module relies on off dataset, we need to convert data first
# focusedIDs: for adaptive purpose, we can choose part of the whole shape
def SparseRefinement(resultSurface=r'noise_shape.off',refinedSurface='refined_shape.off',focusedIDs=range(30), needShapeResult = True, alpha = 50, beta = 0.17):

    global alignedMeshes
    global meshNum
    global landmarkMatrix
    global landmarkNum
    global landmarkList
    
    SetAlignMode(1)
    landmarkNum = numpy.size(focusedIDs)
    SetLandmarkList(focusedIDs)
    n_row = landmarkNum
    n_col = meshNum
    landmarkMatrix = rand(n_row*dimension, n_col)
    for i in range(meshNum): # prepare data matrix
        curMesh = numpy.transpose(alignedMeshes[i])
        for j in range(dimension):
            landmarkMatrix[n_row*j:n_row*j+n_row,i] = list(numpy.take(curMesh[j], landmarkList))
            
    currentShape = ASMtools.LoadOneMesh(resultSurface)
    
    [error_array, refinedLM, refinedShape] = SparseRepresentation([], currentShape, needShapeResult, alpha, beta)
    
    ASMtools.SavePointSet2Off(refinedShape, refinedSurface)
    
# similar to Vector2Mesh in ASMtools, but the data structure of vector is different
# ASMtools.Vector2Mesh: xyzxyz...xyz
# AnomalyDetection.Vector2Mesh: xxx...xyyy...yzzz...z
def Vector2Mesh(vector):
    global dimension
    vecSize = len(vector)
    meshSize = vecSize / dimension
    if dimension == 3:
        result = [[0,0,0]] * meshSize
    elif dimension == 2:
        result = [[0,0]] * meshSize
    for vertexid in range(meshSize):
        if dimension == 3:
            result[vertexid] = [float(vector[vertexid]), float(vector[vertexid+meshSize]), float(vector[vertexid+(2*meshSize)])]
        elif dimension == 2:
            result[vertexid] = [float(vector[vertexid]), float(vector[vertexid+meshSize])]
    return result
    
# convert mesh (point set) to cvxmod vector
def Mesh2Vector(mesh):
    global dimension
    
    meshSize = len(mesh)
    vecSize = meshSize * dimension
    result = rand(vecSize, 1)
    # convert this point set into a vector, which can be used by CVXMOD
    for i in range(len(mesh)):
        for j in range(dimension):
            result[meshSize*j + i] = mesh[i][j]
    return result
